"""Access Bits in a BitVector
"""


class BitField:
    """Data Descriptor for accessing named fields in a BitVector.

    ```python
    > from bitvector import BitVector, BitField
    >
    > class MyBV(BitVector):
    >     byte0 = BitField(0, 8)
    >     byte1 = BitField(8, 8)
    >     bit16 = BitField(16)
    >     bit17 = BitField(17)
    >     pad0  = BitField(18, 4)
    >     byte3 = BitField(32, 8)
    >
    > mybv = MyBV()
    > mybv.byte0 = 0xff
    > mybv.bit17 = 1
    > mybv.byte3 = 0x55
    > mybv
    MyBV(value=0x5502ff, length=128)
    ```

    """

    def __init__(self, offset: int, width: int = 1):
        """
        :param offset: int
        :param width: int
        """
        self.field = slice(offset, offset + width)

    def __get__(self, obj, type=None) -> int:

        value = obj[self.field]
        return value

    def __set__(self, obj, value) -> None:
        prev = obj[self.field]
        obj[self.field] = value

    def __set_name__(self, owner, name) -> None:
        self.name = name



"""A vector of bits for Humans™!
"""

import functools
import operator

from typing import Union
from loguru import logger


@functools.total_ordering
class BitVector:
    """A Bit Vector is a list of bits in packed (integer)
    format that can be accessed by indexing into the vector
    or using a slice (via conventional square brackets 
    notation). 

    """

    @classmethod
    def zeros(cls, size: int = 128):
        return cls(size=size)

    @classmethod
    def ones(cls, size: int = 128):
        bv = cls()
        bv.set()
        return bv

    def __init__(self, value: int = 0, size: int = 128):
        """
        :param value: int
        :param size: int

        Raises:
        - ValueError if size <= 0
        """
        if size <= 0:
            raise ValueError("Size must greater than zero.")

        self.MAX = (1 << size) - 1
        self.value = value

    @property
    def value(self) -> int:
        return getattr(self, "_value", 0)

    @value.setter
    def value(self, new_value: int) -> None:
        self._value = int(new_value) & self.MAX

    def clear(self):
        """Clears all bits in the vector to zero."""
        self._value = 0

    def set(self):
        """Sets all bits in the vector to one."""
        self._value = self.MAX

    def _getb(self, offset: int) -> int:
        """Retrieves the bit value at offset."""

        if offset > len(self) - 1:
            raise IndexError(offset)

        return (self.value >> offset) & 0x1

    def _setb(self, offset: int) -> None:
        """Sets the bit value at offset."""
        if offset > (len(self) - 1):
            raise IndexError(offset)

        self.value |= (1 << offset) & self.MAX

    def _clrb(self, offset: int) -> None:
        """Clears the bit value at offset."""
        self.value &= ~(1 << offset)

    def _setval(self, offset: int, value: int):
        if value:
            self._setb(offset)
        else:
            self._clrb(offset)

    def toggle(self, offset: int) -> int:
        """Toggle the bit at offset in the vector and return the previous value.
        """
        prev = self._getb(offset)
        self.value ^= (1 << offset) & self.MAX
        return prev

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(value={self!s}, size={len(self)})"

    def __str__(self) -> str:
        nibbles = (len(self) // 4) + 1 if (len(self) % 4) else 0
        hexstr = "0x{{:0{0}x}}".format(nibbles)
        return hexstr.format(self.value)

    def __len__(self) -> int:
        """Length of the vector in bits."""
        try:
            return self._len
        except AttributeError:
            pass
        self._len = self.MAX.bit_length()
        return self._len

    def __getitem__(self, key: Union[int, slice]) -> int:
        """Given a key, retrieve a bit or bitfield."""
        try:
            if key < 0:
                key += len(self)
            return self._getb(key)
        except TypeError:
            pass

        try:
            value = 0
            for n, b in enumerate(range(*key.indices(len(self)))):
                v = self._getb(b)
                if v:
                    value += 1 << n
            return value
        except AttributeError:
            raise ValueError(f"Unknown key type: {type(key)}")

    def __setitem__(self, key: Union[int, slice], value: Union[int, bool]) -> None:
        """Given a key, set a bit or bitfield to the supplied value.

        If value is True or False and the key is a slice describing a
        bit field, each bit in the field takes on the value.  Otherwise,
        the value is left shifted and the lsb is added to the next offset
        in the bit field.

        > b[:8] = True   # results in bit0:bit7 == 0b11111111
        > b[:8] = 0x1    # results in bit0:bit7 == 0b00000001
        
        The difference is subtle and perhaps should not be considered a feature.

        Supports negative indexing.
        """
        try:
            if key < 0:
                key += len(self)

            self._setval(key, value)
            return

        except TypeError:
            pass

        try:
            if value is True or value is False:
                for b in range(*key.indices(len(self))):
                    self._setval(b, value)
                return

            for n, b in enumerate(range(*key.indices(len(self)))):
                self._setval(b, (value >> n) & 0x1)

        except AttributeError:
            raise ValueError("Expected int or slice key") from None

    def __binary_op(self, other, func, return_obj: bool = False, reverse: bool = False):
        """Calls the supplied function `func` with self and other.

        If the user sets return_obj to True, a new BitVector initialized with the
        results of `func` is returned. Otherwise the return type is assumed to be
        `int`.  

        :param other: Union[int, BitVector]
        :param func: callable from operator
        :param return_obj: bool
        :return: Union[int, bool, BitVector]
        """

        logger.debug(f"{self} {other} {func} {return_obj} {reverse}")

        try:
            retval = func(self.value, other.value)
            if return_obj:
                size = len(min(self, other, key=len))
                retval = self.__class__(retval, size=size)
            return retval
        except AttributeError:
            pass

        if reverse:
            return func(other, self.value)

        retval = func(self.value, other)

        if return_obj:
            retval = self.__class__(retval)

        return retval

    def __unary_op(self, func, return_obj: bool = False):
        """Calls the supplied function `func` with self and returns the result.

        If return_obj is True, the return value is a BitVector initialized from
        the results of `func`. 

        :param func: callable from operator 
        :param return_obj: bool
        :return: Union[int, BitVector]
        """

        retval = func(self.value) & self.MAX
        if return_obj:
            retval = self.__class__(retval)
        return retval

    def __inplace_op(self, other, func) -> object:
        """Calls the supplied binary function `func` with self and other
        and updates self with the results. 

        :param other:  Union[int, BitVector]
        :param func: Callable from operator
        :return: self
        """
        try:
            self.value = func(self.value, other.value)
        except AttributeError:
            self.value = func(self.value, other)
        return self

    @property
    def bin(self) -> str:
        """Binary string representation of BitVector."""
        return f"0b{bin(self.value)[2:].zfill(len(self))}"

    @property
    def hex(self) -> str:
        """Hexadecimal string representation of BitVector."""
        return hex(self.value)

    @property
    def bytes(self) -> bytes:
        """Byte array representation of BitVector."""
        n = len(self) // 8 + (1 if len(self) % 8 else 0)
        return self.value.to_bytes(n, "big")

    def __bool__(self):
        return bool(self.value)

    __eq__ = functools.partialmethod(__binary_op, func=operator.eq)
    __gt__ = functools.partialmethod(__binary_op, func=operator.gt)

    __add__ = functools.partialmethod(__binary_op, func=operator.add, return_obj=True)
    __radd__ = functools.partialmethod(__binary_op, func=operator.add, reverse=True)
    __iadd__ = functools.partialmethod(__inplace_op, func=operator.add)

    __sub__ = functools.partialmethod(__binary_op, func=operator.sub, return_obj=True)
    __rsub__ = functools.partialmethod(__binary_op, func=operator.sub, reverse=True)
    __isub__ = functools.partialmethod(__inplace_op, func=operator.sub)

    __mul__ = functools.partialmethod(__binary_op, func=operator.mul, return_obj=True)
    __rmul__ = functools.partialmethod(__binary_op, func=operator.mul, reverse=True)
    __imul__ = functools.partialmethod(__inplace_op, func=operator.mul)

    __truediv__ = functools.partialmethod(
        __binary_op, func=operator.truediv, return_obj=True
    )
    __rtruediv__ = functools.partialmethod(
        __binary_op, func=operator.truediv, reverse=True
    )
    __itruediv__ = functools.partialmethod(__inplace_op, func=operator.truediv)

    __floordiv__ = functools.partialmethod(
        __binary_op, func=operator.floordiv, return_obj=True
    )
    __rfloordiv__ = functools.partialmethod(
        __binary_op, func=operator.floordiv, reverse=True
    )
    __ifloordiv__ = functools.partialmethod(__inplace_op, func=operator.floordiv)

    __and__ = functools.partialmethod(__binary_op, func=operator.and_, return_obj=True)
    __rand__ = functools.partialmethod(__binary_op, func=operator.and_, reverse=True)
    __iand__ = functools.partialmethod(__inplace_op, func=operator.and_)

    __or__ = functools.partialmethod(__binary_op, func=operator.or_, return_obj=True)
    __ror__ = functools.partialmethod(__binary_op, func=operator.or_)
    __ior__ = functools.partialmethod(__inplace_op, func=operator.or_)

    __xor__ = functools.partialmethod(__binary_op, func=operator.xor, return_obj=True)
    __rxor__ = functools.partialmethod(__binary_op, func=operator.xor)
    __ixor__ = functools.partialmethod(__inplace_op, func=operator.xor)

    __invert__ = functools.partialmethod(__unary_op, operator.invert, return_obj=True)
    __neg__ = functools.partialmethod(__unary_op, operator.invert, return_obj=True)

    __pos__ = functools.partialmethod(__unary_op, operator.pos, return_obj=True)

    __lshift__ = functools.partialmethod(
        __binary_op, func=operator.lshift, return_obj=True
    )
    __ilshift__ = functools.partialmethod(__inplace_op, func=operator.lshift)

    __rshift__ = functools.partialmethod(
        __binary_op, func=operator.rshift, return_obj=True
    )
    __irshift__ = functools.partialmethod(__inplace_op, func=operator.rshift)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bitvector']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.5.1,<0.6.0']

setup_kwargs = {
    'name': 'bitvector-for-humans',
    'version': '0.11.0',
    'description': 'A simple pure python Bit Vector class for Humans™.',
    'long_description': '# Bit Vectors For Humans™\n\nThis simple bit vector implementation aims to make addressing single\nbits a little less fiddly. It can be used by itself to work with bit\nfields in an integer word, but it really starts to shine when you use\nthe supplied `BitField` descriptor with a subclass of `BitVector`:\n\n```python\n> from bitvector import BitVector, BitField\n>\n> class IOTDeviceCommand(BitVector):\n>     def __init__(self):\n>         super().__init__(size=32)\n>\n>     power = BitField(0, 1) # offset and size\n>     spin  = BitField(1, 1)\n>     speed = BitField(2, 4)\n>     sense = BitField(6, 2)\n>     red   = BitField(8, 8)\n>     blue  = BitField(16, 8)\n>     green = BitField(24, 8)\n>\n> widget_cmd = IOTDeviceCommand()\n> widget_cmd.power = 1\n> widget_cmd.sense = 2\n> widget_cmd.speed = 5\n> widget_cmd.red = 0xaa\n> widget_cmd.blue = 0xbb\n> widget_cmd.green = 0xcc\n> widget_cmd\nIOTDeviceCommand(value=0xccbbaa95, size=32)\n> widget_cmd.bytes\nb\'\\xcc\\xbb\\xaa\\x95\'\n```\n\n\n## Installation\n\n```console\n$ pip install bitvector-for-humans\n$ pydoc bitvector\n...\n```\n\nOr directly from github:\n\n```console\n$ pip install git+https://github.com/JnyJny/bitvector.git\n```\n\n## Motivation\n\n1. Address sub-byte bits in a less error prone way.\n2. Minimize subdependencies.\n3. Learn something about descriptors. \n\n## Caveats\n\nThe tests need expanding and I got lazy when writing the multi-bit\nsetting / getting code and it could undoubtedly be improved. Pull\nrequests gladly accepted.\n\n## Other Ways to Implement a Bit Vector\n\n1. Python builtin `ctypes.Structure` allows sub-byte bit fields\n2. Python builtin `struct` provides extensive support for byte manipulations\n3. Python3 IntEnums can be used to build bit field masks\n4. The plain `int` will serve admirably with bitwise operators\n5. Provide cffi bindings to existing bit-twiddling libraries\n6. Use Numpy bool arrays as the "backing store"\n7. Other good ideas I overlooked, forgot about or just plain don\'t know.\n\n\n\n\n\n\n',
    'author': 'JnyJny',
    'author_email': 'erik.oshaughnessy@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/JnyJny/bitvector.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
